#ifndef ADBMS_DRIVER_H
#define ADBMS_DRIVER_H

#include <stdint.h>
#include "stm32h7xx_hal.h"

#define ERROR_TIME_THRESH 150 // ms

typedef enum {
  SEK_OVERTEMP,
  SEK_UNDERTEMP,
  SEK_OVERVOLT,
  SEK_UNDERVOLT,
  SEK_TOO_FEW_TEMPS,
  SEK_OPENWIRE,
  SEK_EEPROM_ERR,
  SEK_INTERNAL_BMS_TIMEOUT,
  SEK_INTERNAL_BMS_CHECKSUM_FAIL,
  SEK_INTERNAL_BMS_OVERTEMP,
  SEK_INTERNAL_BMS_FAULT,
  SEK_1WIRE_ERR,
  SEK_DS18B20_TIMEOUT,
  NUM_ERROR_KINDS
} SlaveErrorKind;

typedef struct {
  uint8_t data[4];
  uint32_t errors_since;
} SlaveErrorData;

struct ADBMS6830_Internal_Status {
    uint16_t CS_FLT : 16; //ADC fault - mismatch between S- and C-ADC
    uint16_t : 3;
    uint16_t CCTS : 13; //Conversion counter
    uint16_t SMED : 1; //S-ADC multiple trim error (uncorrectable)
    uint16_t SED : 1; //S-ADC single trim error (correctable)
    uint16_t CMED : 1; //C-ADC multiple trim error (uncorrectable)
    uint16_t CED : 1; //C-ADC single trim error (correctable)
    uint16_t VD_UV : 1; //3V digital supply undervoltage
    uint16_t VD_OV : 1; //3V digital supply overvoltage
    uint16_t VA_UV : 1; //5V analog supply undervoltage
    uint16_t VA_OV : 1; //5V analog supply overvoltage
    uint16_t OSCCHK : 1; //Oscillator check
    uint16_t TMODCHK : 1; //Test mode check
    uint16_t THSD : 1; //Thermal shutdown
    uint16_t SLEEP : 1; //Sleep mode previously entered
    uint16_t SPIFLT : 1; //SPI fault
    uint16_t COMPARE : 1; //Comparasion between S- and C-ADC active
    uint16_t VDE : 1; //Supply voltage error
    uint16_t VDEL : 1; //Latent supply voltage error
};

#define MAXIMUM_CELL_VOLTAGES 16
#define MAXIMUM_AUX_VOLTAGES  10
#define MAXIMUM_GPIO          10

typedef struct {
    int16_t cellVoltages[MAXIMUM_CELL_VOLTAGES];
    int16_t auxVoltages[MAXIMUM_AUX_VOLTAGES];

    struct ADBMS6830_Internal_Status status;
    uint16_t internalDieTemp;
    uint16_t analogSupplyVoltage;
    uint16_t digitalSupplyVoltage;
    uint16_t sumOfCellMeasurements;
    uint16_t refVoltage;

    uint16_t GPIO_Values[MAXIMUM_GPIO];

    uint32_t overVoltage;
    uint32_t underVoltage;

} Cell_Module;

extern uint32_t error_sources; // Bitfield of error sources
extern SlaveErrorData error_data[NUM_ERROR_KINDS];

[[gnu::nonnull]]
void AMS_Init(SPI_HandleTypeDef* hspi);

uint8_t AMS_Idle_Loop();

#undef MAXIMUM_CELL_VOLTAGES
#undef MAXIMUM_AUX_VOLTAGES
#undef MAXIMUM_GPIO

#endif // ADBMS_DRIVER_H
