/*
 * AMS_HighLevel.c
 *
 *  Created on: 20.07.2022
 *      Author: max
 */

#include "ADBMS_HighLevel.h"
#include "ADBMS_Abstraction.h"
#include "ADBMS_LL_Driver.h"
#include "config_ADBMS6830.h"
#include "ADBMS_Error.h"
#include "ADBMS_Driver.h"
#include "stm32h7xx_hal.h"
#include "swo_log.h"
#include <stdint.h>
#include <string.h>

Cell_Module modules[N_BMS] = {};
uint32_t balancedCells = 0;
bool balancingActive = false;

uint8_t numberofCells = 15;

uint8_t packetChecksumFails = 0;
#define MAX_PACKET_CHECKSUM_FAILS 5

uint8_t deviceSleeps = 0;
#define MAX_DEVICE_SLEEP 3 //TODO: change to correct value

struct pollingTimes {
  uint32_t S_ADC_OW_CHECK;
  uint32_t TMP1075;
};

struct pollingTimes pollingTimes = {0, 0};

void AMS_Init(SPI_HandleTypeDef* hspi) {
  debug_log(LOG_LEVEL_INFO, "ADBMS6830B HAL - configured for %d controllers and %d cells per controller...", N_BMS, numberofCells);
  if (initAMS(hspi) != HAL_OK) {
    debug_log(LOG_LEVEL_ERROR, "ADBMS6830B HAL - initialization failed");
  }

  pollingTimes = (struct pollingTimes) {HAL_GetTick(), HAL_GetTick()};
}

#define any(x) ({ \
  bool any = false; \
  for (size_t __any_intern_i = 0; __any_intern_i < N_BMS; __any_intern_i++) { \
    Cell_Module module = modules[__any_intern_i]; \
    any |= (x); \
  } \
  any; \
})


uint8_t AMS_Idle_Loop() {
  if (!amsWakeUp()) {
    //error_data.data_kind = SEK_INTERNAL_BMS_TIMEOUT; //we don't receive data for the wakeup command
    //set_error_source(ERROR_SOURCE_INTERNAL);         //so we can't tell if we timed out
  }
  
  packetChecksumFails += amsAuxAndStatusMeasurement(&modules);

  if (any(module.status.SLEEP)) {
    deviceSleeps++;
    if (deviceSleeps > MAX_DEVICE_SLEEP) {
      set_error_source(SEK_INTERNAL_BMS_TIMEOUT);
    } else {
      amsReset();
    }
  }

  if (any(module.status.CS_FLT || module.status.SPIFLT || module.status.CMED || module.status.SMED ||
          module.status.VDE || module.status.VDEL || module.status.OSCCHK || module.status.TMODCHK)) {

      // TODO: handle errors

      // ftcan_marshal_unsigned(&error_data[SEK_INTERNAL_BMS_FAULT].data[0], module.status.CS_FLT, 2);
      /* const uint8_t* ptr = ((uint8_t*)&modules.status) + 4; //skip conversion counter
      error_data[SEK_INTERNAL_BMS_FAULT].data[2] = ptr[1];
      error_data[SEK_INTERNAL_BMS_FAULT].data[3] = ptr[0];
      set_error_source(SEK_INTERNAL_BMS_FAULT); */

      // Fault bits are latched -- clear them so we can check again next
      // iteration.
      amsClearFlag();
  } else {
      clear_error_source(SEK_INTERNAL_BMS_FAULT);
  }

  packetChecksumFails += amsCellMeasurement(&modules);
  packetChecksumFails += amsCheckUnderOverVoltage(&modules);

  if (packetChecksumFails > MAX_PACKET_CHECKSUM_FAILS) {
    set_error_source(SEK_INTERNAL_BMS_CHECKSUM_FAIL);
  }
  
  //TODO: temperature measurement

  bool overvolt = false;
  bool undervolt = false;
  for (size_t i = 0; i < numberofCells; i++) {
    if (any(module.cellVoltages[i] < 2500)) {
      undervolt = true;
      error_data[SEK_UNDERVOLT].data[0] = i;
      uint8_t* ptr = &error_data[SEK_UNDERVOLT].data[1];
      //ftcan_marshal_unsigned(ptr, module.cellVoltages[i], 2);
    } else if (any(module.cellVoltages[i] > 4200)) {
      overvolt = true;
      error_data[SEK_OVERVOLT].data[0] = i;
      uint8_t* ptr = &error_data[SEK_OVERVOLT].data[1];
      //ftcan_marshal_unsigned(ptr, module.cellVoltages[i], 2);
    }
  }

  if (any(module.internalDieTemp > 28000 || module.status.THSD)) { //TODO: change to correct value
    //ftcan_marshal_unsigned(&error_data[SEK_INTERNAL_BMS_OVERTEMP].data[0], module.internalDieTemp, 2);

    set_error_source(SEK_INTERNAL_BMS_OVERTEMP);
  } else {
    clear_error_source(SEK_INTERNAL_BMS_OVERTEMP);
  }

  if (overvolt) {
    set_error_source(SEK_OVERVOLT);
  } else {
    clear_error_source(SEK_OVERVOLT);
  }

  if (undervolt) {
    set_error_source(SEK_UNDERVOLT);
  } else {
    clear_error_source(SEK_UNDERVOLT);
  }

  mcuDelay(10);

  return 0;
}
